/*!
* Copyright Adaptavist 2022 (c) All rights reserved
*/
export class Body {
    constructor(body = null) {
        this.body = body;
    }
    async arrayBuffer() {
        if (this.implementation) {
            if (!this.body) {
                return new ArrayBuffer(0);
            }
            if (typeof this.body === "string") {
                return this.implementation.buffer.encode(this.body);
            }
            else if (typeof this.body['arrayBuffer'] === 'function') {
                const buffer = await this.body.arrayBuffer();
                if (typeof buffer['byteLength'] === 'number') {
                    return buffer;
                }
                else {
                    throw Error('Body is not a known type, only a string, an ArrayBuffer, a Blob or a FormData is supported');
                }
            }
            else if (typeof this.body['entries'] === 'function' && !ArrayBuffer.isView(this.body)) {
                return (await getUint8ArrayOfFormDataEntries(this.implementation, this.body.entries())).uint8Array;
            }
            else if (typeof this.body['byteLength'] === 'number') {
                return this.body ?? new ArrayBuffer(0);
            }
            else {
                throw Error('Body is not a known type, only a string, an ArrayBuffer, a Blob or a FormData is supported');
            }
        }
        else {
            throw Error('Buffer implementation was not specified for Body');
        }
    }
    async text() {
        if (this.implementation) {
            if (!this.body) {
                return '';
            }
            if (typeof this.body === "string") {
                return this.body;
            }
            else if (typeof this.body['text'] === 'function') {
                return this.body.text();
            }
            else if (typeof this.body['entries'] === 'function' && !ArrayBuffer.isView(this.body)) {
                const uint8Arr = (await getUint8ArrayOfFormDataEntries(this.implementation, this.body.entries())).uint8Array;
                return this.implementation.buffer.decode(uint8Arr);
            }
            else if (typeof this.body['byteLength'] === 'number') {
                return this.implementation.buffer.decode(this.body ?? new ArrayBuffer(0));
            }
            else {
                throw Error('Body is not a known type, only a string, an ArrayBuffer, a Blob or a FormData is supported');
            }
        }
        else {
            throw Error('Buffer implementation was not specified for Body');
        }
    }
    async json() {
        const text = await this.text();
        if (text) {
            return JSON.parse(text);
        }
        else {
            throw Error('Cannot parse JSON, body is empty');
        }
    }
    setBufferImplementation(bufferImplementation) {
        this.implementation = bufferImplementation;
    }
}
export class Response extends Body {
    constructor(url, status, statusText, headers, body = null) {
        super(body);
        this.url = url;
        this.status = status;
        this.statusText = statusText;
        this.headers = headers;
        this.ok = this.status >= 200 && this.status < 300;
    }
}
export class Request extends Body {
    constructor(method, url, headers, body) {
        super(body);
        this.method = method;
        this.url = url;
        this.headers = headers;
    }
}
export class Headers {
    constructor(init) {
        this.headers = {};
        if (init) {
            if (init instanceof Headers) {
                const headers = {};
                init.forEach((value, key) => headers[key] = value);
                this.convertHeaders(headers);
            }
            else if (Array.isArray(init)) {
                const headers = init;
                headers.forEach(header => this.headers[header[0]] = header[1].split(',').map(token => token.trim()));
            }
            else {
                this.convertHeaders(init);
            }
            this.sortHeaders(this.headers);
        }
    }
    forEach(callback) {
        for (const entry of this) {
            callback(entry[1], entry[0]);
        }
    }
    append(name, value) {
        let values = this.headers[name];
        if (!values) {
            values = [];
            this.headers[name] = values;
        }
        values.push(value);
        this.sortHeaders(this.headers);
    }
    delete(name) {
        delete this.headers[name];
    }
    get(name) {
        const values = this.headers[name];
        if (values) {
            return values.join(', ');
        }
        return null;
    }
    getAll(name) {
        return this.headers[name] || [];
    }
    has(name) {
        return !!this.headers[name];
    }
    raw() {
        return this.headers;
    }
    set(name, value) {
        this.headers[name] = [value];
        this.sortHeaders(this.headers);
    }
    entries() {
        let pointer = 0;
        const keys = Object.keys(this.headers);
        const total = keys.length;
        return {
            next: () => {
                if (pointer < total) {
                    const key = keys[pointer++];
                    return {
                        done: false,
                        value: [key, this.headers[key].join(', ')]
                    };
                }
                else {
                    return {
                        done: true,
                        value: null
                    };
                }
            }
        };
    }
    keys() {
        let pointer = 0;
        const keys = Object.keys(this.headers);
        const total = keys.length;
        return {
            next: () => {
                if (pointer < total) {
                    return {
                        done: false,
                        value: keys[pointer++]
                    };
                }
                else {
                    return {
                        done: true,
                        value: null
                    };
                }
            }
        };
    }
    values() {
        let pointer = 0;
        const keys = Object.keys(this.headers);
        const total = keys.length;
        return {
            next: () => {
                if (pointer < total) {
                    return {
                        done: false,
                        value: this.headers[keys[pointer++]].join(', ')
                    };
                }
                else {
                    return {
                        done: true,
                        value: null
                    };
                }
            }
        };
    }
    sortHeaders(headers) {
        this.headers = {};
        Object.keys(headers).sort().forEach(key => this.headers[key] = headers[key]);
    }
    convertHeaders(headers) {
        for (const [key, value] of Object.entries(headers)) {
            this.headers[key] = value.split(',').map(token => token.trim());
        }
    }
    [Symbol.iterator]() {
        return this.entries();
    }
}
// Modified version of https://github.com/jimmywarting/FormData/blob/master/FormData.js
// TODO: complete TS conversion, and maybe rework as it looks like a piece of crap implementation
export class FormData {
    constructor() {
        this._data = [];
    }
    /**
     * Append a field
     *
     * @param   {string}           name      field name
     * @param   {string|Blob|File} value     string / blob / file
     * @param   {string=}          filename  filename to use with blob
     * @return  {undefined}
     */
    // @ts-ignore
    append(name, value, filename) {
        this.ensureArgs(arguments, 2);
        // @ts-ignore
        var [a, s, d] = this.normalizeArgs.apply(null, arguments);
        // @ts-ignore
        this._data.push([a, s, d]);
    }
    /**
     * Delete all fields values given name
     *
     * @param   {string}  name  Field name
     * @return  {undefined}
     */
    // @ts-ignore
    delete(name) {
        this.ensureArgs(arguments, 1);
        // @ts-ignore
        const res = [];
        name = String(name);
        // @ts-ignore
        this.each(this._data, entry => {
            if (entry[0] !== name) {
                // @ts-ignore
                res.push(entry);
            }
        });
        // @ts-ignore
        this._data = res;
    }
    /**
     * Iterate over all fields as [name, value]
     *
     * @return {Iterator}
     */
    *entries() {
        for (var i = 0; i < this._data.length; i++) {
            yield this.normalizeValue(this._data[i]);
        }
    }
    /**
     * Iterate over all fields
     *
     * @param   {Function}  callback  Executed for each item with parameters (value, name, thisArg)
     * @param   {Object=}   thisArg   `this` context for callback function
     * @return  {undefined}
     */
    // @ts-ignore
    forEach(callback, thisArg) {
        this.ensureArgs(arguments, 1);
        for (const [name, value] of this) {
            callback.call(thisArg, value, name, this);
        }
    }
    /**
     * Return first field value given name
     * or null if non existen
     *
     * @param   {string}  name      Field name
     * @return  {string|File|null}  value Fields value
     */
    // @ts-ignore
    get(name) {
        this.ensureArgs(arguments, 1);
        const entries = this._data;
        name = String(name);
        for (let i = 0; i < entries.length; i++) {
            if (entries[i][0] === name) {
                return this.normalizeValue(this._data[i])[1];
            }
        }
        return null;
    }
    /**
     * Return all fields values given name
     *
     * @param   {string}  name  Fields name
     * @return  {Array}         [{String|File}]
     */
    // @ts-ignore
    getAll(name) {
        this.ensureArgs(arguments, 1);
        const result = [];
        name = String(name);
        for (let i = 0; i < this._data.length; i++) {
            if (this._data[i][0] === name) {
                // @ts-ignore
                result.push(this.normalizeValue(this._data[i])[1]);
            }
        }
        return result;
    }
    /**
     * Check for field name existence
     *
     * @param   {string}   name  Field name
     * @return  {boolean}
     */
    // @ts-ignore
    has(name) {
        this.ensureArgs(arguments, 1);
        name = String(name);
        for (let i = 0; i < this._data.length; i++) {
            if (this._data[i][0] === name) {
                return true;
            }
        }
        return false;
    }
    /**
     * Iterate over all fields name
     *
     * @return {Iterator}
     */
    *keys() {
        for (const [name] of this) {
            yield name;
        }
    }
    /**
     * Overwrite all values given name
     *
     * @param   {string}    name      Filed name
     * @param   {string}    value     Field value
     * @param   {string=}   filename  Filename (optional)
     * @return  {undefined}
     */
    // @ts-ignore
    set(name, value, filename) {
        this.ensureArgs(arguments, 2);
        name = String(name);
        const result = [];
        let replaced = false;
        for (let i = 0; i < this._data.length; i++) {
            const match = this._data[i][0] === name;
            if (match) {
                if (!replaced) {
                    // @ts-ignore
                    result[i] = this.normalizeArgs.apply(null, arguments);
                    replaced = true;
                }
            }
            else {
                result.push(this._data[i]);
            }
        }
        if (!replaced) {
            // @ts-ignore
            result.push(this.normalizeArgs.apply(null, arguments));
        }
        // @ts-ignore
        this._data = result;
    }
    /**
     * Iterate over all fields
     *
     * @return {Iterator}
     */
    *values() {
        for (const [, value] of this) {
            yield value;
        }
    }
    getBlob() {
        const boundary = '----formdata-' + Math.random();
        const chunks = [];
        // @ts-ignore
        const getFileName = (fileName) => fileName ? `; filename="${fileName}"` : '';
        for (const [name, value, fileName] of this) {
            // @ts-ignore
            chunks.push(`--${boundary}\r\n`);
            if (value instanceof Blob) {
                chunks.push(
                // @ts-ignore
                `Content-Disposition: form-data; name="${name}"; filename="${fileName || value.name}"\r\n`, `Content-Type: ${value.type || 'application/octet-stream'}\r\n\r\n`, value, '\r\n');
            }
            else {
                chunks.push(
                // @ts-ignore
                `Content-Disposition: form-data; name="${name}"${getFileName(fileName)}\r\n\r\n${value}\r\n`);
            }
        }
        // @ts-ignore
        chunks.push(`--${boundary}--`);
        return new Blob(this.implementation, chunks, {
            type: 'multipart/form-data; boundary=' + boundary
        });
    }
    /**
     * The class itself is iterable
     * alias for formdata.entries()
     *
     * @return  {Iterator}
     */
    [Symbol.iterator]() {
        return this.entries();
    }
    /**
     * Create the default string description.
     *
     * @return  {string} [object FormData]
     */
    toString() {
        return '[object FormData]';
    }
    // @ts-ignore
    ensureArgs(args, expected) {
        if (args.length < expected) {
            throw new TypeError(`${expected} argument required, but only ${args.length} present.`);
        }
    }
    // @ts-ignore
    normalizeArgs(name, value, filename) {
        return value instanceof Blob
            // normalize name and filename if adding an attachment
            ? [String(name), value, filename !== undefined
                    ? filename + '' // Cast filename to string if 3th arg isn't undefined
                    : typeof value.name === 'string' // if name prop exist
                        ? value.name // Use File.name
                        : 'blob'] // otherwise fallback to Blob
            // If no attachment, just cast the args to strings
            : [String(name), String(value), filename];
    }
    // @ts-ignore
    normalizeValue([name, value, filename]) {
        if (value instanceof Blob) {
            // Should always returns a new File instance
            // TODO: maybe add File implementation?
            // value = new File([value], filename, {
            //     type: value.type,
            //     lastModified: (value as any).lastModified
            // })
        }
        return [name, value, filename];
    }
    // @ts-ignore
    each(arr, cb) {
        for (let i = 0; i < arr.length; i++) {
            cb(arr[i]);
        }
    }
    setBufferImplementation(bufferImplementation) {
        this.implementation = bufferImplementation;
    }
}
// Modified version of: https://github.com/bjornstar/blob-polyfill/blob/master/Blob.js
// TODO: complete TS conversion, and maybe rework as it looks like a piece of crap implementation
export class Blob {
    // @ts-ignore
    constructor(implementation, chunks, opts) {
        this.implementation = implementation;
        chunks = chunks || [];
        opts = opts == null ? {} : opts;
        for (var i = 0, len = chunks.length; i < len; i++) {
            var chunk = chunks[i];
            if (chunk instanceof Blob) {
                chunks[i] = chunk._buffer;
            }
            else if (typeof chunk === 'string') {
                if (!this.implementation) {
                    throw Error('Platform specific implementation is missing');
                }
                chunks[i] = this.implementation?.buffer.encode(chunk);
            }
            else if (ArrayBuffer.prototype.isPrototypeOf(chunk) || this.isArrayBufferView(chunk)) {
                chunks[i] = this.bufferClone(chunk);
            }
            else if (this.isDataView(chunk)) {
                chunks[i] = this.bufferClone(chunk.buffer);
            }
            else {
                if (!this.implementation) {
                    throw Error('Platform specific implementation is missing');
                }
                chunks[i] = this.implementation?.buffer.encode(String(chunk));
            }
        }
        // @ts-ignore
        this._buffer = global.Uint8Array
            ? this.concatTypedArrays(chunks)
            : [].concat.apply([], chunks);
        this.size = this._buffer.length;
        this.type = opts.type || '';
        if (/[^\u0020-\u007E]/.test(this.type)) {
            this.type = "";
        }
        else {
            this.type = this.type.toLowerCase();
        }
    }
    arrayBuffer() {
        return Promise.resolve(this._buffer);
    }
    arrayBufferSync() {
        return this._buffer;
    }
    text() {
        if (!this.implementation) {
            throw Error('Platform specific implementation is missing');
        }
        return Promise.resolve(this.implementation?.buffer.decode(this._buffer));
    }
    textSync() {
        if (!this.implementation) {
            throw Error('Platform specific implementation is missing');
        }
        return this.implementation?.buffer.decode(this._buffer);
    }
    slice(start, end, type) {
        const slice = this._buffer.slice(start || 0, end || this._buffer.length);
        return new Blob(this.implementation, [slice], { type });
    }
    // @ts-ignore
    isArrayBufferView(obj) {
        return ArrayBuffer.isView(obj);
    }
    // @ts-ignore
    bufferClone(buf) {
        var view = new Array(buf.byteLength);
        var array = new Uint8Array(buf);
        var i = view.length;
        while (i--) {
            view[i] = array[i];
        }
        return view;
    }
    // @ts-ignore
    isDataView(obj) {
        return obj && DataView.prototype.isPrototypeOf(obj);
    }
    // @ts-ignore
    concatTypedArrays(chunks) {
        var size = 0;
        var j = chunks.length;
        while (j--) {
            size += chunks[j].length;
        }
        var b = new Uint8Array(size);
        var offset = 0;
        for (var i = 0; i < chunks.length; i++) {
            var chunk = chunks[i];
            b.set(chunk, offset);
            offset += chunk.byteLength || chunk.length;
        }
        return b;
    }
}
export class BaseErrorStrategyBuilder {
    constructor(options) {
        this.options = options;
        if (!options) {
            this.options = {};
        }
    }
    /**
     * Error handler for HTTP code 400 (Bad Request). Return the value you wish to continue with, or return 'retry()' for retrying the request, or return 'continuePropagation()' for allowing the error to propagate.
     */
    http400Error(handler) {
        this.options.handleHttp400Error = handler;
        return this.buildNewBuilder(this.options);
    }
    /**
     * Error handler for HTTP code 401 (Unauthorized). Return the value you wish to continue with, or return 'retry()' for retrying the request, or return 'continuePropagation()' for allowing the error to propagate.
     */
    http401Error(handler) {
        this.options.handleHttp401Error = handler;
        return this.buildNewBuilder(this.options);
    }
    /**
     * Error handler for HTTP code 403 (Forbidden). Return the value you wish to continue with, or return 'retry()' for retrying the request, or return 'continuePropagation()' for allowing the error to propagate.
     */
    http403Error(handler) {
        this.options.handleHttp403Error = handler;
        return this.buildNewBuilder(this.options);
    }
    /**
     * Error handler for HTTP code 404 (Not Found). Return the value you wish to continue with, or return 'retry()' for retrying the request, or return 'continuePropagation()' for allowing the error to propagate.
     */
    http404Error(handler) {
        this.options.handleHttp404Error = handler;
        return this.buildNewBuilder(this.options);
    }
    /**
     * Error handler for HTTP code 429 (Too Many Requests). Return the value you wish to continue with, or return 'retry()' for retrying the request, or return 'continuePropagation()' for allowing the error to propagate.
     */
    http429Error(handler) {
        this.options.handleHttp429Error = handler;
        return this.buildNewBuilder(this.options);
    }
    /**
     * Error handler for any HTTP code in 500 range (various server side errors). Return the value you wish to continue with, or return 'retry()' for retrying the request, or return 'continuePropagation()' for allowing the error to propagate.
     */
    http5xxError(handler) {
        this.options.handleHttp5xxError = handler;
        return this.buildNewBuilder(this.options);
    }
    /**
     * Error handler for any invalid HTTP code (anything other than in 200 and 300 range). Return the value you wish to continue with, or return 'retry()' for retrying the request, or return 'continuePropagation()' for allowing the error to propagate.
     */
    httpAnyError(handler) {
        this.options.handleHttpAnyError = handler;
        return this.buildNewBuilder(this.options);
    }
    /**
     * Error handler for unexpected error, usually either network error or JSON parsing error. Return the value you wish to continue with, or return 'retry()' for retrying the request, or return 'continuePropagation()' for allowing the error to propagate.
     */
    unexpectedError(handler) {
        this.options.handleUnexpectedError = handler;
        return this.buildNewBuilder(this.options);
    }
    /**
     * Error handler for any error. Return the value you wish to continue with, or return 'retry()' for retrying the request, or return 'continuePropagation()' for allowing the error to propagate.
     */
    anyError(handler) {
        this.options.handleAnyError = handler;
        return this.buildNewBuilder(this.options);
    }
    /**
     * Retry when HTTP request is being rate limited (429 response).
     * @param total How many times to re-try. Set to -1 to retry indefinitely (default).
     * @param verbose Whether to show rate limit warning message. Default: true.
     * @param delay How long to delay before re-trying again, in milliseconds. Default: 1 second.
     * @param delayIncrease Delay to increase per each unsuccessful re-try, in milliseconds. Default: 1 second.
     */
    retryOnRateLimiting(total = -1, delay = 1000, delayIncrease = 1000, verbose = true) {
        return this.http429Error(getRetryErrorHandler(total, delay, delayIncrease, verbose));
    }
    /**
     * Retry on any HTTP error.
     * @param total How many times to re-try. Set to -1 to retry indefinitely. Default: 3.
     * @param delay How long to delay before re-trying again, in milliseconds. Set to -1 to try to follow 'Retry-After' header value. Default: 1 second.
     * @param delayIncrease Delay to increase per each unsuccessful re-try, in milliseconds. Default: 1 second.
     */
    retryOnHttpError(total = 3, delay = 1000, delayIncrease = 1000) {
        return this.httpAnyError(getRetryErrorHandler(total, delay, delayIncrease));
    }
    /**
     * Retry on any failure.
     * @param total How many times to re-try. Set to -1 to retry indefinitely. Default: 3.
     * @param delay How long to delay before re-trying again, in milliseconds. Set to -1 to try to follow 'Retry-After' header value. Default: 1 second.
     * @param delayIncrease Delay to increase per each unsuccessful re-try, in milliseconds. Default: 1 second.
     */
    retryOnAnyError(total = 3, delay = 1000, delayIncrease = 1000) {
        return this.anyError(getRetryErrorHandler(total, delay, delayIncrease));
    }
    build() {
        return this.options;
    }
}
export class RetryRequested {
    constructor(delay) {
        this.delay = delay;
    }
}
export class PropagationRequested {
    constructor(skipErrorStrategyPropagation) {
        this.skipErrorStrategyPropagation = skipErrorStrategyPropagation;
    }
}
export class HttpError extends Error {
    constructor(response, message, name = 'HttpError') {
        super(message);
        this.response = response;
        this.name = name;
    }
}
export class UnauthorizedError extends HttpError {
    constructor(response, message) {
        super(response, message, 'UnauthorizedError');
    }
}
export class ForbiddenError extends HttpError {
    constructor(response, message) {
        super(response, message, 'ForbiddenError');
    }
}
export class NotFoundError extends HttpError {
    constructor(response, message) {
        super(response, message, 'NotFoundError');
    }
}
export class BadRequestError extends HttpError {
    constructor(response, message) {
        super(response, message, 'BadRequestError');
    }
}
export class ServerError extends HttpError {
    constructor(response, message) {
        super(response, message, 'ServerError');
    }
}
export class TooManyRequestsError extends HttpError {
    constructor(response, message) {
        super(response, message, 'TooManyRequestsError');
    }
}
export class UnexpectedError extends Error {
    constructor(originalError, message) {
        super(message);
        this.originalError = originalError;
        this.message = message;
        this.name = 'UnexpectedError';
    }
}
export function constructUrl(url, queryParams, duplicateParams = false) {
    if (!queryParams) {
        return url;
    }
    let i = 0;
    for (const key in queryParams) {
        let value = queryParams[key];
        if (value !== undefined && value !== null) {
            if (Array.isArray(value)) {
                value = duplicateParams ? value.join(`&${encodeURIComponent(key)}=`) : value.map(v => encodeURIComponent(v)).join(',');
                url += `${(i == 0) ? '?' : '&'}${encodeURIComponent(key)}=${value}`;
            }
            else {
                url += `${(i == 0) ? '?' : '&'}${encodeURIComponent(key)}=${encodeURIComponent(value)}`;
            }
            i++;
        }
    }
    return url;
}
export function getHeaders(body, manualHeaders) {
    let headers = manualHeaders ? new Headers(manualHeaders) : new Headers();
    if (!headers.get('Accept')) {
        headers.append('Accept', 'application/json');
    }
    if (body && !headers.get('Content-Type')) {
        if (typeof body['arrayBuffer'] === 'function' && !!body.type) {
            const type = body.type;
            headers.append('Content-Type', type);
        }
        else {
            headers.append('Content-Type', 'application/json; charset=utf-8');
        }
    }
    return headers;
}
export async function executeHttpCall(implementation, request, errorStrategyBuilderConstructor, globalErrorHandlers, errorHandlers, errorMessageBuilder, customErrorHandlers = [], skipBodyParsing = false, attempt = 1) {
    try {
        const response = await implementation.performHttpCall(request);
        for (const customErrorHandler of customErrorHandlers) {
            await customErrorHandler.processResponse?.(response);
        }
        if (response.ok) {
            if (skipBodyParsing) {
                if (typeof skipBodyParsing === 'function') {
                    return await skipBodyParsing(response);
                }
                return await response.text();
            }
            return await response.json();
        }
        else {
            throw await mapStatusCodeToError(implementation, response, errorMessageBuilder);
        }
    }
    catch (originalError) {
        const processedErrorHandlers = getErrorHandlers(errorStrategyBuilderConstructor, globalErrorHandlers, errorHandlers);
        const verboseMessage = new VerboseMessageBuilder(request.method, request.url);
        try {
            return await handleError(originalError, attempt, verboseMessage, processedErrorHandlers, customErrorHandlers);
        }
        catch (error) {
            if (error instanceof RetryRequested) {
                await delay(error.delay);
                return await executeHttpCall(implementation, request, errorStrategyBuilderConstructor, globalErrorHandlers, errorHandlers, errorMessageBuilder, customErrorHandlers, skipBodyParsing, ++attempt);
            }
            if (error instanceof PropagationRequested) {
                throw originalError;
            }
            throw error;
        }
    }
}
function getErrorHandlers(errorStrategyBuilderConstructor, globalErrorHandlers, errorHandlers) {
    if (globalErrorHandlers && typeof globalErrorHandlers === 'function') {
        const response = globalErrorHandlers(errorStrategyBuilderConstructor());
        if (response instanceof BaseErrorStrategyBuilder) {
            globalErrorHandlers = response.build();
        }
        else if ('build' in response && typeof response.build === 'function') {
            globalErrorHandlers = response.build();
        }
        else {
            throw Error(`Unknown global error strategy builder response type: ${response}`);
        }
    }
    if (errorHandlers && typeof errorHandlers === 'function') {
        const response = errorHandlers(errorStrategyBuilderConstructor());
        if (response instanceof BaseErrorStrategyBuilder) {
            errorHandlers = response.build();
        }
        else {
            throw Error(`Unknown error strategy builder response type: ${response}`);
        }
    }
    return { ...(globalErrorHandlers ?? {}), ...(errorHandlers ?? {}) };
}
async function handleError(e, attempt, verboseMessage, errorHandlers, customErrorHandlers = []) {
    let error;
    if (!(e instanceof HttpError || customErrorHandlers.some(handler => e.name === handler.errorName))) {
        error = new UnexpectedError(e, e.toString ? e.toString() : 'Unexpected Error');
    }
    else {
        error = e;
    }
    let handlerName = undefined;
    try {
        if (errorHandlers) {
            const { handleHttp400Error, handleHttp401Error, handleHttp403Error, handleHttp404Error, handleHttp429Error, handleHttp5xxError, handleHttpAnyError, handleUnexpectedError, handleAnyError } = errorHandlers;
            if (error instanceof HttpError) {
                if (error instanceof BadRequestError && handleHttp400Error) {
                    handlerName = 'handleHttp400Error';
                    return await handleErrorStrategyResponse(handleHttp400Error, error, attempt);
                }
                if (error instanceof UnauthorizedError && handleHttp401Error) {
                    handlerName = 'handleHttp401Error';
                    return await handleErrorStrategyResponse(handleHttp401Error, error, attempt);
                }
                if (error instanceof ForbiddenError && handleHttp403Error) {
                    handlerName = 'handleHttp403Error';
                    return await handleErrorStrategyResponse(handleHttp403Error, error, attempt);
                }
                if (error instanceof NotFoundError && handleHttp404Error) {
                    handlerName = 'handleHttp404Error';
                    return await handleErrorStrategyResponse(handleHttp404Error, error, attempt);
                }
                if (error instanceof TooManyRequestsError && handleHttp429Error) {
                    handlerName = 'handleHttp429Error';
                    return await handleErrorStrategyResponse(handleHttp429Error, error, attempt, verboseMessage);
                }
                if (error instanceof ServerError && handleHttp5xxError) {
                    handlerName = 'handleHttp5xxError';
                    return await handleErrorStrategyResponse(handleHttp5xxError, error, attempt);
                }
                if (handleHttpAnyError) {
                    handlerName = 'handleHttpAnyError';
                    return await handleErrorStrategyResponse(handleHttpAnyError, error, attempt);
                }
            }
            if (error instanceof UnexpectedError && handleUnexpectedError) {
                handlerName = 'handleUnexpectedError';
                return await handleErrorStrategyResponse(handleUnexpectedError, error, attempt);
            }
            for (const customErrorHandler of customErrorHandlers) {
                if (error.name === customErrorHandler.errorName && typeof errorHandlers[customErrorHandler.handlerName] === 'function') {
                    handlerName = customErrorHandler.handlerName;
                    return await handleErrorStrategyResponse(errorHandlers[customErrorHandler.handlerName], error, attempt);
                }
            }
            if (handleAnyError) {
                handlerName = 'handleAnyError';
                return await handleErrorStrategyResponse(handleAnyError, error, attempt);
            }
        }
    }
    catch (e) {
        if (e instanceof RetryRequested) {
            throw e;
        }
        if (e instanceof PropagationRequested) {
            if (e.skipErrorStrategyPropagation) {
                throw e;
            }
            if (handlerName && errorHandlers) {
                // Delete the current callback method so when a propagation is allowed to progress then this error handler cannot be reached any longer
                delete errorHandlers[handlerName];
                return await handleError(error, attempt, verboseMessage, errorHandlers);
            }
        }
        throw e;
    }
    throw error;
}
async function handleErrorStrategyResponse(callback, error, attempt, verboseMessage) {
    const response = callback(error, attempt, verboseMessage);
    if (response instanceof RetryRequested) {
        throw response;
    }
    if (response instanceof PropagationRequested) {
        throw response;
    }
    return response;
}
export async function mapStatusCodeToError(implementation, response, errorMessageBuilder) {
    const fetchId = response.headers.get('x-stitch-fetch-id');
    let errorMessage = `Unexpected HTTP response: ${response.status}${fetchId ? ` - Fetch ID: ${fetchId}` : ''}`;
    if (implementation.errorMessageBuilder) {
        errorMessage = (await implementation.errorMessageBuilder(response, errorMessage)) ?? errorMessage;
    }
    if (errorMessageBuilder) {
        errorMessage = (await errorMessageBuilder(response, errorMessage)) ?? errorMessage;
    }
    switch (response.status) {
        case 400:
            return new BadRequestError(response, errorMessage);
        case 401:
            return new UnauthorizedError(response, errorMessage);
        case 403:
            return new ForbiddenError(response, errorMessage);
        case 404:
            return new NotFoundError(response, errorMessage);
        case 429:
            return new TooManyRequestsError(response, errorMessage);
        case 500:
        case 501:
        case 502:
        case 503:
        case 504:
        case 505:
            return new ServerError(response, errorMessage);
        default:
            return new HttpError(response, errorMessage);
    }
}
export function delay(ms) {
    return new Promise(resolve => {
        if (ms === 0) {
            resolve();
        }
        else {
            setTimeout(() => resolve(), ms);
        }
    });
}
/**
 * Asks the current request to be re-tried.
 * @param delay Delay in milliseconds after which the re-try is invoked. Optional, defaults to 0.
 */
export function retry(delay = 0) {
    return new RetryRequested(delay);
}
/**
 * Asks the current error handler to propagate the error to next error handler.
 * @param skipErrorStrategyPropagation If this value is set to true then any further error strategy handler is skipped and the error is sent directly to Promise rejection. Optional, defaults to false.
 */
export function continuePropagation(skipErrorStrategyPropagation = false) {
    return new PropagationRequested(skipErrorStrategyPropagation);
}
/**
 * Returns error handler for re-trying on failures.
 * @param total How many times to re-try. Set to -1 to retry indefinitely (default).
 * @param delay How long to delay before re-trying again, in milliseconds. Set to -1 to try to follow 'Retry-After' header value (default).
 * @param delayIncrease Delay to increase per each unsuccessful re-try, in milliseconds. Default: 0 second.
 * @param verbose Whether to show a specific warning message.
 */
export function getRetryErrorHandler(total = -1, delay = -1, delayIncrease = 0, verbose) {
    return (error, attempt, verboseMessage) => {
        if (delay === -1) {
            try {
                const retryAfter = error.response.headers.get('Retry-After');
                if (retryAfter) {
                    delay = parseInt(retryAfter) * 1000;
                }
                else {
                    throw new Error('Retry-After header is not present in response');
                }
            }
            catch (e) {
                delay = 1000;
            }
        }
        if (error instanceof TooManyRequestsError) {
            const stitchRateLimit = error.response.headers.get('x-stitch-rate-limit');
            if (verbose && verboseMessage && (attempt <= total || total === -1)) {
                console.warn(verboseMessage.getTooManyRequestsVerboseMessage(delay + ((attempt - 1) * delayIncrease), Boolean(stitchRateLimit)));
            }
        }
        return attempt <= total || total === -1 ? retry(delay + ((attempt - 1) * delayIncrease)) : continuePropagation();
    };
}
export function constructHeaders(body, manualHeaders) {
    return getHeaders(body, manualHeaders);
}
export class BaseApiCore {
    constructor() {
        this.implementation = this.getPlatformImplementation();
    }
    async fetch(resource, options) {
        let request;
        if (resource instanceof Request) {
            request = resource;
        }
        else if (options?.body && typeof options?.body['entries'] === 'function' && !ArrayBuffer.isView(options.body)) {
            const { boundary, uint8Array } = await getUint8ArrayOfFormDataEntries(this.implementation, options.body.entries());
            const formDataHeader = { 'Content-Type': `multipart/form-data; boundary=${boundary}` };
            request = new Request(options?.method ?? 'POST', resource, constructHeaders(options?.body, { ...formDataHeader, ...options?.headers }), uint8Array.buffer);
        }
        else {
            request = new Request(options?.method ?? 'GET', resource, constructHeaders(options?.body, options?.headers), options?.body);
        }
        request.setBufferImplementation(this.implementation);
        return await this.implementation.performHttpCall(request);
    }
}
export class VerboseMessageBuilder {
    constructor(method, url) {
        this.method = method;
        this.url = url;
    }
    getTooManyRequestsVerboseMessage(delayMs, headerApplied) {
        return headerApplied
            ? `API call to ${this.method} ${this.url} was rate limited by ScriptRunner Connect, retrying again in ${delayMs / 1000} seconds. Please consider upgrading your plan to increase the API calls rate limit`
            : `API call to ${this.method} ${this.url} was rate limited, retrying again in ${delayMs / 1000} seconds`;
    }
}
export async function getUint8ArrayOfFormDataEntries(implementation, formEntries) {
    const boundary = '----formdata-' + Math.random();
    const chunks = [];
    const getFileName = (fileName) => fileName ? `; filename="${fileName}"` : '';
    for (const [name, value, fileName] of formEntries) {
        chunks.push(`--${boundary}\r\n`);
        if (typeof value.arrayBuffer === 'function') {
            const arrayBufferFromBlob = await value.arrayBuffer();
            chunks.push(`Content-Disposition: form-data; name="${name}"; filename="${fileName || value.name}"\r\n`, `Content-Type: ${value.type || 'application/octet-stream'}\r\n\r\n`, arrayBufferFromBlob, '\r\n');
        }
        else {
            chunks.push(`Content-Disposition: form-data; name="${name}"${getFileName(fileName)}\r\n\r\n${value}\r\n`);
        }
    }
    chunks.push(`--${boundary}--`);
    const encodedChunks = chunks.map(chunk => typeof chunk === 'string' ? implementation?.buffer.encode(chunk) : chunk);
    const totalByteLength = encodedChunks.reduce((total, chunk) => total + chunk.byteLength, 0);
    const uint8Array = new Uint8Array(totalByteLength);
    let offset = 0;
    for (const chunk of encodedChunks) {
        uint8Array.set(new Uint8Array(chunk), offset);
        offset += chunk.byteLength;
    }
    return {
        uint8Array,
        boundary
    };
}
//# sourceMappingURL=index.js.map